# Measure the time it takes to execute the file
start_time_model <- Sys.time()

# Load Model ############################################################
#This script is called from UK to set model data:
contador <- 0
empty_model <- NULL
# We download the model by days
for (date in dates) {
  conc_path_day  <- paste0(date, "00")
  
  file           <- paste0(name_pollutant, "_", conc_path_day, ".nc", sep="")
  file_path      <- paste0(model_path, file)
  
  if (file.exists(file_path)) {
    netcdf_model      <- RNetCDF::open.nc(file_path) 

    #Extraiem les dades del model
    if(substr(date,1,4)==2017 & name_pollutant %in% c("pm10","pm2p5")){
      name_pollutant_var <- paste0(name_pollutant,"cmaq")
    }else{
      name_pollutant_var <- name_pollutant
    }
    
    model_day <- RNetCDF::var.get.nc(netcdf_model, name_pollutant_var) 
    
#FIXME
    # if(date==dates[1] | contador==1){
    if(!exists("n_lat") || !exists("n_lon")){
      n_lat     <- dim(model_day)[1]
      n_lon     <- dim(model_day)[2]
      model     <- array(dim=c(n_lat,n_lon,0))
    }
    
    if(model_time_resolution=="hourly"){
      model_day      <- model_day[,,1:24]
    }else if(model_time_resolution=="daily"){
      model_day      <- model_day[,] #Meriem: revisar
    } else{
      print("Wrong time resolution")
    }
    
    # model data is in ppb, we want to have ug/m3
    # excepte per CO, que ja ens va be en ppm
    if(name_pollutant!="co"){
      model_day       <- model_day*mod_constant 
    }
    
    if(model_time_resolution=="hourly" & time_resolution=="daily"){
      averaged_array  <- apply(model_day, c(1, 2), mean) # Matrix with the means of the day
      model_day       <- array(averaged_array, 
                               dim = c(nrow(averaged_array), ncol(averaged_array), 1)) # Arming the 3rd dimension (time)
    } else if( model_time_resolution=="daily" & time_resolution=="hourly"){
      print(paste0("WARNING: time resolution of the model and other data (observations mostly), do not match"))
    }
    
  }else{
      print(paste0("WARNING: There is no model or file on: ", date))
      contador   <- contador + 1
      empty_model <- c(empty_model,date)
      
  #IP
      # if (!exists("n_lat") || !exists("n_lon")) {
      #   n_lat <- 397
      #   n_lon <- 397
      #   model <- array(dim=c(n_lat,n_lon,0))
      # }
      
      if (!exists("n_lat") || !exists("n_lon")) {
        model <- NULL
        model_day <- NULL
      }else{

        #if we dont have data for a certain day its filled by NaN
        if(time_resolution=="hourly"){
          model_day       <- array(NaN, dim=c(n_lat,n_lon,24)) 
        }else if (time_resolution=="daily"){
          model_day       <- array(NaN, dim=c(n_lat,n_lon,1))
        }else{
          print("Wrong model_time_resolution")
        }
      }
    }
  model            <- abind(model,model_day,along=3)
}

# if we had days missing model from the first day of the year, we will have gaps
# we fix that by filling from the top with NANs
n_gaps <- length(dates) - dim(model)[3]  #days with gaps from the 1st day
if(time_resolution=="hourly"){
  n_gaps <- n_gaps*24
}  #hours with gaps from the 1st day
if(n_gaps>0){
  model_gap_top <- array(NaN, dim=c(n_lat,n_lon,n_gaps))
  
  model <- abind(model_gap_top, model, along=3)
  rm(model_gap_top)
}


print(paste0("The number of missing raw model (CALIOPE) is: ", contador))
if(!is.null(empty_model)){
  print("These days are: ")
  print(empty_model)
}

lat_model      <- RNetCDF::var.get.nc(netcdf_model, 'lat') #Latitud
lat_model      <- sapply(lat_model, as.numeric)            #era una large matrix, ara es una large numeric
lon_model      <- RNetCDF::var.get.nc(netcdf_model, 'lon') #Longitud
lon_model      <- sapply(lon_model, as.numeric)

coords_model   <- cbind(lon_model, lat_model)  
  


#extract the dates in which we have model
available_dates_model <- dates[!dates %in% empty_model]

if(time_resolution=="hourly"){
  available_dates_model <- rep(available_dates_model, 24) %>%
    sort() %>%
    paste(seq(0,23)) %>% 
    as.POSIXct(format="%Y%m%d %H")
}



if (save_model){
  saveRDS(object=model, file=paste0(save_result,"model_",file_format, "_",pollutant,".rds"))
  saveRDS(object=coords_model, file=paste0(save_result,"coords_model_",file_format, "_",pollutant,".rds"))
  print("Model saved")
} else {
  print("Model have not been saved")
}

end_time_model    <- Sys.time()
temps_total_model <- end_time_model - start_time_model
print(paste0("local model, execution time :",temps_total_model, "s" ))
