# Measure the time it takes to execute the file
start_time_UK_LOOCV <- Sys.time()
  

#LOOCV_total and model_raw_at_AQMS is created empty to fill it later with LOOCV and raw model data respectivelly

if (execute_uk) {
  #UK_total_df           <- list(length=length_loop) #we will have a large list with a data.frame for each hour/day
  UK_total_df           <- list() #we will have a large list with a data.frame for each hour/day
}
if (execute_loocv) {
  LOOCV_total           <- matrix(NA,ncol=nrow(local_stations@data),nrow=0) #list(length=nrow(date_times)) #we will have a large list with a data.frame for each hour
  colnames(LOOCV_total) <- local_stations$ref_obs #we want a matrix with empty data and the name of stations
}

if (execute_uk=="FALSE" && execute_loocv=="FALSE") {
  print("The user did not choose any method to apply")
}

model_raw_at_AQMS <- model_raw_at_AQMS_df




# Lists to save the slope and the interception between observations and model:

# slope_v  <-  array(NA,dim = length(available_dates_model))
# inter_v  <-  array(NA,dim = length(available_dates_model))
# r_v      <-  array(NA,dim = length(available_dates_model))
slope_v  <-  array(NA,dim = length_loop)
inter_v  <-  array(NA,dim = length_loop)
r_v      <-  array(NA,dim = length_loop)


# Loop UK - LOOCV ############################################################
# date_0 <- df_uk_observations[1,"date"] %>%
#   lubridate::date()
date_0 <- df_uk_observations[1,"date"] %>%
  lubridate::date()

count_i <- 0

print("Loop UK - LOOCV")
#starting OK and LOOCV:
# for(i in 1:length(available_dates_model)){
# for(i in 1:length(df_uk_observations$date)){
for (i in 1:length_loop) {
  # i <- 10
  # date_i <- available_dates_model[i]

  # date_i <- dates[i]
  
  if(time_resolution=="daily"){
    date_i <- dates[i]
    date_i <- as.Date(date_i,format="%Y%m%d",tz="UTC")
  }else if(time_resolution=="hourly"){
    date_i <- date_times[i,1]
    date_i <- as.POSIXct(date_i,format="%Y%m%d %H",tz="UTC")
  }

  count_i <- count_i+1
  
  print(paste0("loop UK or LOOCV: ", i, " (",date_i,")"))
  
  #extract observations hourly
  df_obs_hour <- subset(df_uk_observations[i,], select = -c(date))
  

  # #To avoid what we suppose wrong observations, we will subtitute the values higher than limit microg/m3
  # Get the row and column indices of values that exceed the limit
  indices <- which(df_obs_hour > limit, arr.ind = TRUE)
  
  # Check if any values exceed the limit
  if (nrow(indices) > 0) {
    # Modify the values that exceed the limit
    # df_obs_hour[indices] <- limit
    df_obs_hour[indices] <- NA
    
    # Print a message
    message(paste0("Observations exceeding the limit ", limit, " microg/m3 (mg/m3 for CO) have been ignored. The number of observations modified are: ", nrow(indices), " in day/hour: ", i))
  }
  
  #raw model:
  df_model_uk <- model[,,i]
  
  if (!any(is.na(df_model_uk))){
    df_model_uk                  <- data.frame(lat=lat_model,lon=lon_model,pollutant_model=sapply(df_model_uk,as.matrix))
    coordinates(df_model_uk)     <- ~lon+lat
    sp::proj4string(df_model_uk) <- crs_latlon
    
    #config the coordinates in UTM
    if (compute_in_UTM) {
      df_model_uk <- spTransform(df_model_uk, sp::CRS(crs_utm))
    }
    
    
    # ok_hour <- model_raw_at_AQMS[i,-which(names(model_raw_at_AQMS)=="date")]
    
    
    ok_hour <- model_raw_at_AQMS %>% 
      dplyr::filter(date==date_i) %>%
      dplyr::select(-date)
    
    
    df_loocv <- data.frame(lat=local_stations@coords[,2],
                           lon=local_stations@coords[,1],
                           pollutant_model=as.numeric(ok_hour[1,]),
                           ref_obs=local_stations@data$ref_obs) 
        
    df_obs_loocv <- data.frame(pollutant_obs = as.numeric(df_obs_hour[1,]), 
                               ref_obs = colnames(df_obs_hour))

    # remove prefixes and sufixes of ghost data station names
    if(file_format=="ghost"){
      df_loocv$ref_obs <- df_loocv$ref_obs %>%
        gsub("STA_","",.) %>% gsub("\\_.*","",.) %>% gsub("STA-","",.)
      df_obs_loocv$ref_obs <- df_obs_loocv$ref_obs %>%
        gsub("STA_","",.) %>% gsub("\\_.*","",.) %>% gsub("STA-","",.)
    }
    
    
    llista_loocv <- merge(df_loocv, df_obs_loocv, by = "ref_obs")
    llista_loocv <- na.omit(llista_loocv) 

    if (!all(is.na(llista_loocv$pollutant_obs))){
      
      
      coordinates(llista_loocv)        <- ~lon+lat
      sp::proj4string(llista_loocv)    <- crs_latlon
      
      #config the coordinates in UTM
      if (compute_in_UTM) {
        llista_loocv <- spTransform(llista_loocv, sp::CRS(crs_utm))
      } else {
        sp::proj4string(llista_loocv) <- crs_latlon
      }
      
      # Compute LOOCV variogram
      
      pollutant.fit_loocv <- automap::autofitVariogram(formula=pollutant_obs~pollutant_model,input_data=llista_loocv, verbose=FALSE)
      
      #Emergency Variogram if range of the current Variogram is negative
      if(any(pollutant.fit_loocv$var_model$range<0)){
        print('UK Variogram had negative range')
        pollutant.fit_loocv=readRDS(variogram_uk)
      }
      
      #pollutant.fit_loocv$var_model[2,"range"] <- pollutant.fit_loocv$var_model[2,"range"]
      
      #coef(lm(pollutant_obs ~  pollutant_model , llista_uk))['pollutant_model'] is the same as slope:
      slope <- summary(lm(pollutant_obs ~  pollutant_model , llista_loocv))$coefficients[,1][-1]
      interc <- ((lm(pollutant_obs ~  pollutant_model , llista_loocv))$coefficients)[1]
      r <- cor(llista_loocv$pollutant_obs,llista_loocv$pollutant_model, method = "pearson")
      #p     <-summary(lm(pollutant_obs ~  pollutant_model , llista_uk))$coefficients[,4][-1] #p value.
      
      #plot(llista_loocv$pollutant_model,llista_loocv$pollutant_obs)
      
      slope_v[i]  <-  slope
      inter_v[i]  <- interc
      r_v[i]      <- r
      
# TODO filtre mes restrictiu
      if(slope[1] < -20){
        
        if(execute_uk){
          # UK
          uk_interpolation              <- as.data.frame(df_model_uk$pollutant_model)
          colnames(uk_interpolation)    <- c("var1.pred")
          coordinates(uk_interpolation) <- df_model_uk@coords
          
          #config the coordinates in UTM
          if (compute_in_UTM) {
            sp::proj4string(uk_interpolation) <- crs_utm
          } else {
            sp::proj4string(uk_interpolation) <- crs_latlon
          }
        }
        
        
        if(execute_loocv){
          # LOOCV
          #loocv_interpolation              <- as.data.frame(llista_loocv$pollutant_model)
          #colnames(loocv_interpolation)    <- c("var1.pred")
          loocv_interpolation              <- data.frame(var1.pred= array(NA,dim=length(llista_loocv$pollutant_model) ) )
          coordinates(loocv_interpolation) <- llista_loocv@coords
          
          #config the coordinates in UTM
          if (compute_in_UTM) {
            sp::proj4string(loocv_interpolation) <- crs_utm
          } else {
            sp::proj4string(loocv_interpolation) <- crs_latlon
          }
        }
        
        print(paste0("The slope in ", date_i, " is ", slope, ". We did not apply LOOCV, raw-model value is picked for this step"))
        
      }else {
        
        #Save the first variogram executed for a certain pollutant if we already don't have one saved:
        if (!file.exists(variogram_uk)) {
          saveRDS(pollutant.fit_loocv,variogram_uk)
        }
        
        if(execute_uk){
          # UK
          uk_gstat          <- gstat(formula=pollutant_obs~pollutant_model,data=llista_loocv, model = pollutant.fit_loocv$var_model)
          uk_interpolation  <- predict(object=uk_gstat, newdata=df_model_uk)
        }
        
        if(execute_loocv){
          # LOOCV
          loocv_interpolation  <- krige.cv(pollutant_obs~pollutant_model,llista_loocv,pollutant.fit_loocv$var_model)
        }
      }
      
      if(execute_uk){
        UK_total_df[[count_i]]       <- uk_interpolation@data$var1.pred
      }
      if(execute_loocv){
        uk_loocv_h             <- matrix(data=loocv_interpolation@data$var1.pred, nrow=1)
        colnames(uk_loocv_h)   <- llista_loocv@data$ref_obs
        LOOCV_total            <- plyr::rbind.fill(as.data.frame(LOOCV_total),as.data.frame(uk_loocv_h))
      }
      
    } else{
      print(paste0("WARNING: The observations are corrupted or non-existent in: ", date_i))
      
      slope_v[i]  <- NA
      inter_v[i]  <- NA
      r_v[i]      <- NA
      
      
      if(execute_uk){
        UK_total_df[[count_i]]       <- rep(NA, length(df_model_uk))
      }
      if(execute_loocv){
        
        uk_loocv_h             <- matrix(NA, nrow = 1, ncol = nrow(local_stations))
        colnames(uk_loocv_h )  <- local_stations@data$ref_obs
        LOOCV_total            <- plyr::rbind.fill(as.data.frame(LOOCV_total),as.data.frame(uk_loocv_h))
      }
    }
    
  # Si no es pot fer el UK, es guarda amb tot NA
  } else{ 
    print(paste0("WARNING: The model is corrupted or non-existent in: ", date_i))
    if(execute_uk){
      UK_total_df[[count_i]]       <- rep(NA, length(df_model_uk))
    }
    if(execute_loocv){ #M: no m'acaba de convèncer
    
      uk_loocv_h             <- matrix(NA, nrow = 1, ncol = nrow(local_stations))
      colnames(uk_loocv_h )  <- local_stations@data$ref_obs
      LOOCV_total            <- plyr::rbind.fill(as.data.frame(LOOCV_total),as.data.frame(uk_loocv_h))
    }
    
  }
  #extract date of the following hour
  # if(i != length(available_dates_model)){
  if(i != length_loop){
      date_i <- df_uk_observations[i+1,"date"] %>%
      lubridate::date()
    
    # check if this is the last iteration before changing to a new day
    if(date_i > date_0){
      print(paste0("saving ", date_0))
      saveRDS(UK_total_df,
              file = paste0(save_result,"UK_",file_format,"_",pollutant,
                            if (execute_uk_gbm) "_gbm_" else "_",
                            strftime(date_0,"%Y%m%d") ,".rds"))
      
      # save as csv (rows are model cells, cols are days/hours)
      # do.call(rbind, UK_total_df) %>% 
      #   t() %>%
      #   as.data.frame() %>%
      #   write.table(file=paste0(save_result,"UK_",file_format,"_",pollutant, 
      #                         if (execute_uk_gbm) "_gbm_" else "_",
      #                         strftime(date_0,"%Y%m%d") ,".csv"),
      #             sep=",",row.names=F,col.names=F)
      
      date_0 <- date_i
      count_i <- 0
      
      #restart object
      UK_total_df <- list()
    }
  #save the last day 
  }else{
    print("last day... saving")
    saveRDS(UK_total_df,
            file = paste0(save_result,"UK_",file_format,"_",pollutant,
                          if (execute_uk_gbm) "_gbm_" else "_",
                          strftime(date_0,"%Y%m%d") ,".rds"))
    
    # save as csv (rows are model cells, cols are days/hours)
    # do.call(rbind, UK_total_df) %>% 
    #   t() %>%
    #   as.data.frame() %>%
    #   write.table(file=paste0(save_result,"UK_",file_format,"_",pollutant, 
    #                         if (execute_uk_gbm) "_gbm_" else "_",
    #                         strftime(date_0,"%Y%m%d") ,".csv"),
    #               sep=",",row.names=F,col.names=F)
    
  }
  
}


# Set LOOCV_total_df with dates
if(execute_loocv){
  LOOCV_total_df        <- as.data.frame(LOOCV_total)
  for(i in 1:nrow(local_stations)){
    LOOCV_total_df[[i]] <- replace(LOOCV_total_df[[i]], LOOCV_total_df[[i]]<0, 0)
  } 
  
  if(time_resolution=="hourly"){
    #UK_LOOCV_total_df <- cbind(dates,UK_LOOCV_total_df_wd)
    LOOCV_total_df        <- cbind(date = date_times$date, LOOCV_total_df)
  }else if(time_resolution=="daily"){
    LOOCV_total_df        <- cbind(date = dates, LOOCV_total_df)
    # LOOCV_total_df        <- cbind(date = available_dates_model, LOOCV_total_df)
    LOOCV_total_df$date <- as.Date(LOOCV_total_df$date, format = "%Y%m%d")
    
  }else{
    print("Wrong time_resolution")
  }
}
  


if (save_LOOCV){
  # if(execute_uk){
  #   saveRDS(object=UK_total_df, file = paste0(save_result, "UK_", file_format, "_", pollutant, if (execute_uk_gbm) "_gbm_" else "_", "df.rds")) 
  #   print("UK saved")
  # }
  if(execute_loocv){
    saveRDS(object=LOOCV_total_df, file = paste0(save_result, "LOOCV_", file_format, "_", pollutant, if (execute_uk_gbm) "_gbm_" else "_", "df.rds"))
    print("LOOCV saved")
  }
  
} else {
  print("LOOCV - UK have not been saved")
}

end_time_UK_LOOCV     <- Sys.time()
temps_total_UK_LOOCV  <- end_time_UK_LOOCV  - start_time_UK_LOOCV 
print(paste0("UK or LOOCV, execution time:",temps_total_UK_LOOCV))
